
; FASM SOURCE FILE
; http://flatassembler.net

; <SI>mple <C>ompiler
; native functions (complex number trigonometric functions)

; Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
; All rights reserved
;
; http://maalchemist.ru
; http://maalchemist.narod.ru
; maa@maalchemist.ru
; maalchemist@yandex.ru
; maalchemist@gmail.com

; ------------------------------------------------------------------------------

; complex sine
; z = a + ib
; csin(a,b)
; csin(z) = sin(a)*cosh(b) + icos(a)*sinh(b)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'
;
; a' = sin(a)*cosh(b)
; b' = cos(a)*sinh(b)

macro @@csin.sse
{
$usprocf @@csin

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        movapd  xmm3, xmm1                      ; xmm3 = b
        mov     rdx, @@fsincos.enter
        call    rdx                             ; xmm0 = sin(a)  | xmm1 = cos(a)
        @@.xch  xmm3, xmm0                      ; xmm0 = b       | xmm3 = sin(a)
        movapd  xmm2, xmm1                      ; xmm2 = cos(a)
        mov     rdx, @@fchsh.enter
        call    rdx                             ; xmm0 = cosh(b) | xmm1 = sinh(b)
        mulsd   xmm0, xmm3                      ; xmm0 = sin(a)*cosh(b)
        mulsd   xmm1, xmm2                      ; xmm1 = cos(a)*sinh(b)
   .return:
        retn
$endp

macro @@.csin.fpu
\{
$usprocf @@csin

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b      | st1 = a
        ;
        mov     rdx, @@shch.enter.fpu
        call    rdx                             ; st0 = ch(b)  | st1 = sh(b)  | st2 = a
        fxch    st2                             ; st0 = a      | st1 = sh(b)  | st2 = ch(b)
        @@st0.sin.cos.core.reduce
        fsincos                                 ; st0 = cos(a) | st1 = sin(a) | st2 = sh(b) | st3 = ch(b)
        fmulp   st2, st0                        ; st0 = sin(a) | st1 = cos(a)sh(b) | st2 = ch(b)
        fmulp   st2, st0                        ; st0 = cos(a)sh(b) | st1 = sin(a)ch(b)
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
\}

$usprocf @@csin.enter.fpu
                                                ; st0 = b      | st1 = a
        mov     rdx, @@shch.enter.fpu
        call    rdx                             ; st0 = ch(b)  | st1 = sh(b)  | st2 = a
        fxch    st2                             ; st0 = a      | st1 = sh(b)  | st2 = ch(b)
        @@st0.sin.cos.core.reduce
        fsincos                                 ; st0 = cos(a) | st1 = sin(a) | st2 = sh(b) | st3 = ch(b)
        fmulp   st2, st0                        ; st0 = sin(a) | st1 = cos(a)sh(b) | st2 = ch(b)
        fmulp   st2, st0                        ; st0 = cos(a)sh(b) | st1 = sin(a)ch(b)
   .return:
        retn
$endp
}

macro @@csin.fpu
{
$usprocf @@csin

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b      | st1 = a
        mov     rdx, @@shch.enter
        call    rdx                             ; st0 = ch(b)  | st1 = sh(b)  | st2 = a
        fxch    st2                             ; st0 = a      | st1 = sh(b)  | st2 = ch(b)
        @@st0.sin.cos.core.reduce
        fsincos                                 ; st0 = cos(a) | st1 = sin(a) | st2 = sh(b) | st3 = ch(b)
        fmulp   st2, st0                        ; st0 = sin(a) | st1 = cos(a)sh(b) | st2 = ch(b)
        fmulp   st2, st0                        ; st0 = cos(a)sh(b) | st1 = sin(a)ch(b)
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@csin.sse
$else
  @@csin.fpu
$end

; ------------------------------------------------------------------------------

; complex cosine
; z = a + ib
; ccos(a,b)
; ccos(z) = cos(a)*cosh(b) - isin(a)*sinh(b)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'
;
; a' =  cos(a)*cosh(b)
; b' = -sin(a)*sinh(b)

macro @@ccos.sse
{
$usprocf @@ccos

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        movapd  xmm3, xmm1                      ; xmm3 = b
        mov     rdx, @@fsincos.enter
        call    rdx                             ; xmm0 = sin(a)  | xmm1 = cos(a)
        @@.xch  xmm3, xmm0                      ; xmm0 = b       | xmm3 = sin(a)
        movapd  xmm2, xmm1                      ; xmm2 = cos(a)
        mov     rdx, @@fchsh.enter
        call    rdx                             ; xmm0 = cosh(b) | xmm1 = sinh(b)
        mulsd   xmm0, xmm2                      ; xmm0 = cos(a)*cosh(b)
        mulsd   xmm1, xmm3                      ; xmm1 = sin(a)*sinh(b)
        @@.chs  xmm1                            ; xmm1 = -sin(a)*sinh(b)
   .return:
        retn
$endp

macro @@.ccos.fpu
\{
$usprocf @@ccos

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b      | st1 = a
        ;
        mov     rdx, @@shch.enter.fpu
        call    rdx                             ; st0 = ch(b)  | st1 = sh(b)  | st2 = a
        fxch    st2                             ; st0 = a      | st1 = sh(b)  | st2 = ch(b)
        @@st0.sin.cos.core.reduce
        fsincos                                 ; st0 = cos(a) | st1 = sin(a) | st2 = sh(b) | st3 = ch(b)
        fmulp   st3, st0                        ; st0 = sin(a) | st1 = sh(b)  | st2 = cos(a)ch(b)
        fchs                                    ; st0 =-sin(a) | st1 = sh(b)  | st3 = cos(a)ch(b)
        fmulp                                   ; st0 =-sin(a)sh(b) | st1 = cos(a)ch(b)
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
\}

$asprocf @@ccos.enter.fpu
                                                ; st0 = b      | st1 = a
        mov     rdx, @@shch.enter.fpu
        call    rdx                             ; st0 = ch(b)  | st1 = sh(b)  | st2 = a
        fxch    st2                             ; st0 = a      | st1 = sh(b)  | st2 = ch(b)
        @@st0.sin.cos.core.reduce
        fsincos                                 ; st0 = cos(a) | st1 = sin(a) | st2 = sh(b) | st3 = ch(b)
        fmulp   st3, st0                        ; st0 = sin(a) | st1 = sh(b)  | st2 = cos(a)ch(b)
        fchs                                    ; st0 =-sin(a) | st1 = sh(b)  | st3 = cos(a)ch(b)
        fmulp                                   ; st0 =-sin(a)sh(b) | st1 = cos(a)ch(b)
   .return:
        retn
$endp
}

macro @@ccos.fpu
{
$usprocf @@ccos

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b      | st1 = a
        mov     rdx, @@shch.enter
        call    rdx                             ; st0 = ch(b)  | st1 = sh(b)  | st2 = a
        fxch    st2                             ; st0 = a      | st1 = sh(b)  | st2 = ch(b)
        @@st0.sin.cos.core.reduce
        fsincos                                 ; st0 = cos(a) | st1 = sin(a) | st2 = sh(b) | st3 = ch(b)
        fmulp   st3, st0                        ; st0 = sin(a) | st1 = sh(b)  | st2 = cos(a)ch(b)
        fchs                                    ; st0 =-sin(a) | st1 = sh(b)  | st3 = cos(a)ch(b)
        fmulp                                   ; st0 =-sin(a)sh(b) | st1 = cos(a)ch(b)
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@ccos.sse
$else
  @@ccos.fpu
$end

; ------------------------------------------------------------------------------

; complex secant
; z = a + ib
; csec(a,b) = 1/ccos(a,b)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@csec.sse
{
$usprocf @@csec

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        mov     rdx, @@ccos.enter
        call    rdx
        mov     rdx, @@cinv.enter
        call    rdx
   .return:
        retn
$endp

macro @@.csec.fpu
\{
$usprocf @@csec

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        mov     rdx, @@ccos.enter.fpu
        call    rdx
        mov     rdx, @@cinv.enter.fpu
        call    rdx
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
\}
}

macro @@csec.fpu
{
$usprocf @@csec

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        mov     rdx, @@ccos.enter
        call    rdx
        mov     rdx, @@cinv.enter
        call    rdx
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@csec.sse
$else
  @@csec.fpu
$end

; ------------------------------------------------------------------------------

; complex cosecant
; z = a + ib
; ccosec(a,b) = 1/csin(a,b)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@ccosec.sse
{
$usprocf @@ccosec

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        mov     rdx, @@csin.enter
        call    rdx
        mov     rdx, @@cinv.enter
        call    rdx
   .return:
        retn
$endp

macro @@.ccosec.fpu
\{
$usprocf @@ccosec

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        mov     rdx, @@csin.enter.fpu
        call    rdx
        mov     rdx, @@cinv.enter.fpu
        call    rdx
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
\}
}

macro @@ccosec.fpu
{
$usprocf @@ccosec

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        mov     rdx, @@csin.enter
        call    rdx
        mov     rdx, @@cinv.enter
        call    rdx
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@ccosec.sse
$else
  @@ccosec.fpu
$end

; ------------------------------------------------------------------------------

; complex tangent
; z = a + ib
; ctan(a,b) = csin(a,b) / ccos(a,b)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@ctan.sse
{
$usprocf @@ctan

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        movapd  xmm6, xmm0
        movapd  xmm7, xmm1
        mov     rdx, @@ccos.enter
        call    rdx
        @@.xch  xmm6, xmm0
        @@.xch  xmm7, xmm1
        mov     rdx, @@csin.enter
        call    rdx
        movapd  xmm2, xmm6
        movapd  xmm3, xmm7
        mov     rdx, @@cdiv.enter
        call    rdx
   .return:
        retn
$endp

macro @@.ctanh.fpu
\{
$usprocf @@ctan

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b      | st1 = a
        ;
        fld     st1
        fld     st1                             ; st0 = b      | st1 = a      | st2 = b      | st3 = a
        mov     rdx, @@ccos.enter.fpu
        call    rdx                             ; st0 = ccos.b | st1 = ccos.a | st2 = b      | st3 = a
        fxch    st2                             ; st0 = b      | st1 = ccos.a | st2 = ccos.b | st3 = a
        fxch                                    ; st0 = ccos.a | st1 = b      | st2 = ccos.b | st3 = a
        fxch    st3                             ; st0 = a      | st1 = b      | st2 = ccos.b | st3 = ccos.a
        fxch                                    ; st0 = b      | st1 = a      | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@csin.enter.fpu
        call    rdx                             ; st0 = csin.b | st1 = csin.a | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@cdivr.enter.fpu
        call    rdx
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
\}
}

macro @@ctan.fpu
{
$usprocf @@ctan

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b      | st1 = a
        fld     st1
        fld     st1                             ; st0 = b      | st1 = a      | st2 = b      | st3 = a
        mov     rdx, @@ccos.enter
        call    rdx                             ; st0 = ccos.b | st1 = ccos.a | st2 = b      | st3 = a
        fxch    st2                             ; st0 = b      | st1 = ccos.a | st2 = ccos.b | st3 = a
        fxch                                    ; st0 = ccos.a | st1 = b      | st2 = ccos.b | st3 = a
        fxch    st3                             ; st0 = a      | st1 = b      | st2 = ccos.b | st3 = ccos.a
        fxch                                    ; st0 = b      | st1 = a      | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@csin.enter
        call    rdx                             ; st0 = csin.b | st1 = csin.a | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@cdivr.enter
        call    rdx
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@ctan.sse
$else
  @@ctan.fpu
$end

; ------------------------------------------------------------------------------

; complex cotangent
; z = a + ib
; ccotan(a,b) = ccos(a,b) / csin(a,b)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@ccotan.sse
{
$usprocf @@ccotan

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        movapd  xmm6, xmm0
        movapd  xmm7, xmm1
        mov     rdx, @@csin.enter
        call    rdx
        @@.xch  xmm6, xmm0
        @@.xch  xmm7, xmm1
        mov     rdx, @@ccos.enter
        call    rdx
        movapd  xmm2, xmm6
        movapd  xmm3, xmm7
        mov     rdx, @@cdiv.enter
        call    rdx
   .return:
        retn
$endp

macro @@.ccotanh.fpu
\{
$usprocf @@ccotan

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b      | st1 = a
        ;
        fld     st1
        fld     st1                             ; st0 = b      | st1 = a      | st2 = b      | st3 = a
        mov     rdx, @@ccos.enter.fpu
        call    rdx                             ; st0 = ccos.b | st1 = ccos.a | st2 = b      | st3 = a
        fxch    st2                             ; st0 = b      | st1 = ccos.a | st2 = ccos.b | st3 = a
        fxch                                    ; st0 = ccos.a | st1 = b      | st2 = ccos.b | st3 = a
        fxch    st3                             ; st0 = a      | st1 = b      | st2 = ccos.b | st3 = ccos.a
        fxch                                    ; st0 = b      | st1 = a      | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@csin.enter.fpu
        call    rdx                             ; st0 = csin.b | st1 = csin.a | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@cdiv.enter.fpu
        call    rdx
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
\}
}

macro @@ccotan.fpu
{
$usprocf @@ccotan

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b      | st1 = a
        fld     st1
        fld     st1                             ; st0 = b      | st1 = a      | st2 = b      | st3 = a
        mov     rdx, @@ccos.enter
        call    rdx                             ; st0 = ccos.b | st1 = ccos.a | st2 = b      | st3 = a
        fxch    st2                             ; st0 = b      | st1 = ccos.a | st2 = ccos.b | st3 = a
        fxch                                    ; st0 = ccos.a | st1 = b      | st2 = ccos.b | st3 = a
        fxch    st3                             ; st0 = a      | st1 = b      | st2 = ccos.b | st3 = ccos.a
        fxch                                    ; st0 = b      | st1 = a      | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@csin.enter
        call    rdx                             ; st0 = csin.b | st1 = csin.a | st2 = ccos.b | st3 = ccos.a
        mov     rdx, @@cdiv.enter
        call    rdx
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@ccotan.sse
$else
  @@ccotan.fpu
$end

; ------------------------------------------------------------------------------
; ------------------------------------------------------------------------------

; inverse complex sine
; z = a + ib
; casin(z)
; -i*ln(i*(z + (z^2 - 1)^(1/2)))
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@casin.sse.fpu
{
$usprocf @@casin

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter.fpu
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter.fpu
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter.fpu
        call    rdx                             ; z+(z^2-1)^(1/2)
        fchs
        fxch                                    ; i*(z+(z^2-1)^(1/2))
        mov     rdx, @@cln.enter.fpu
        call    rdx                             ; ln(i*(z+(z^2-1)^(1/2)))
        fxch
        fchs                                    ; (-i)*ln(i*(z+(z^2-1)^(1/2)))
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
}

macro @@casin.fpu
{
$usprocf @@casin

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter
        call    rdx                             ; z+(z^2-1)^(1/2)
        fchs
        fxch                                    ; i*(z+(z^2-1)^(1/2))
        mov     rdx, @@cln.enter
        call    rdx                             ; ln(i*(z+(z^2-1)^(1/2)))
        fxch
        fchs                                    ; (-i)*ln(i*(z+(z^2-1)^(1/2)))
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@casin.sse.fpu
$else
  @@casin.fpu
$end

; ------------------------------------------------------------------------------

; inverse complex cosine
; z = a + ib
; cacos(z)
; -i*ln(z + (z^2 - 1)^(1/2))
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@cacos.sse.fpu
{
$usprocf @@cacos

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter.fpu
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter.fpu
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter.fpu
        call    rdx                             ; z+(z^2-1)^(1/2)
        mov     rdx, @@cln.enter.fpu
        call    rdx                             ; ln(z+(z^2-1)^(1/2))
        fxch
        fchs                                    ; (-i)*ln(z+(z^2-1)^(1/2))
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
}

macro @@cacos.fpu
{
$usprocf @@cacos

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter
        call    rdx                             ; z+(z^2-1)^(1/2)
        mov     rdx, @@cln.enter
        call    rdx                             ; ln(z+(z^2-1)^(1/2))
        fxch
        fchs                                    ; (-i)*ln(z+(z^2-1)^(1/2))
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@cacos.sse.fpu
$else
  @@cacos.fpu
$end

; ------------------------------------------------------------------------------

; inverse complex secant
; z = a + ib
; casec(z) = cacos(1/z)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@casec.sse.fpu
{
$usprocf @@casec

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        mov     rdx, @@cinv.enter.fpu
        call    rdx                             ; 1/z
   .cacos:
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter.fpu
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter.fpu
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter.fpu
        call    rdx                             ; z+(z^2-1)^(1/2)
        mov     rdx, @@cln.enter.fpu
        call    rdx                             ; ln(z+(z^2-1)^(1/2))
        fxch
        fchs                                    ; (-i)*ln(z+(z^2-1)^(1/2))
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
}

macro @@casec.fpu
{
$usprocf @@casec

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        mov     rdx, @@cinv.enter
        call    rdx                             ; 1/z
   .cacos:
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter
        call    rdx                             ; z+(z^2-1)^(1/2)
        mov     rdx, @@cln.enter
        call    rdx                             ; ln(z+(z^2-1)^(1/2))
        fxch
        fchs                                    ; (-i)*ln(z+(z^2-1)^(1/2))
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@casec.sse.fpu
$else
  @@casec.fpu
$end

; ------------------------------------------------------------------------------

; inverse complex cosecant
; z = a + ib
; cacosec(z) = casin(1/z)
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@cacosec.sse.fpu
{
$usprocf @@cacosec

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        mov     rdx, @@cinv.enter.fpu
        call    rdx                             ; 1/z
   .casin:
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter.fpu
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter.fpu
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter.fpu
        call    rdx                             ; z+(z^2-1)^(1/2)
        fchs
        fxch                                    ; i*(z+(z^2-1)^(1/2))
        mov     rdx, @@cln.enter.fpu
        call    rdx                             ; ln(i*(z+(z^2-1)^(1/2)))
        fxch
        fchs                                    ; (-i)*ln(i*(z+(z^2-1)^(1/2)))
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
}

macro @@cacosec.fpu
{
$usprocf @@cacosec

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        mov     rdx, @@cinv.enter
        call    rdx                             ; 1/z
   .casin:
        fld     st1
        fld     st1                             ; st0 = b | st1 = a | st2 = b | st3 = a
        mov     rdx, @@cpow2.enter
        call    rdx                             ; st0:st1 = z^2
        fld1
        fsubp   st2, st0                        ; z^2-1
        mov     rdx, @@croot2.enter
        call    rdx                             ; (z^2-1)^(1/2)
        mov     rdx, @@cadd.enter
        call    rdx                             ; z+(z^2-1)^(1/2)
        fchs
        fxch                                    ; i*(z+(z^2-1)^(1/2))
        mov     rdx, @@cln.enter
        call    rdx                             ; ln(i*(z+(z^2-1)^(1/2)))
        fxch
        fchs                                    ; (-i)*ln(i*(z+(z^2-1)^(1/2)))
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@cacosec.sse.fpu
$else
  @@cacosec.fpu
$end

; ------------------------------------------------------------------------------

; inverse complex tangent
; z = a + ib
; catan(z)
; (1/2)*(-i)*ln((1+i*z)/(1-i*z))
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@catan.sse.fpu
{
$usprocf @@catan

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        fchs
        fxch                                    ; i*z = (ia-b)
                                                ; st0 = a | st1 = -b
        fld     st1
        fld     st1
        fchs                                    ; st0:st1 = ~i*z | st2:st3 = i*z
                                                ; ~i*z = (-ia-b)
        fld1                                    ; st0 = 1 | st1:st2 = ~i*z | st3:st4 = i*z
        fadd    st4, st0
        fsubrp  st2, st0                        ; st0:st1 = 1-i*z | st2:st3 = 1+i*z
        mov     rdx, @@cdiv.enter.fpu
        call    rdx                             ; (1+i*z)/(1-i*z)
        mov     rdx, @@cln.enter.fpu
        call    rdx                             ; ln((1+i*z)/(1-i*z))
        fxch
        fchs                                    ; (-i)*ln((1+i*z)/(1-i*z))
        fldq    ??.0.5
        fmul    st2, st0
        fmulp                                   ; (1/2)*(-i)*ln((1+i*z)/(1-i*z))
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
}

macro @@catan.fpu
{
$usprocf @@catan

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        fchs
        fxch                                    ; i*z = (ia-b)
                                                ; st0 = a | st1 = -b
        fld     st1
        fld     st1
        fchs                                    ; st0:st1 = ~i*z | st2:st3 = i*z
                                                ; ~i*z = (-ia-b)
        fld1                                    ; st0 = 1 | st1:st2 = ~i*z | st3:st4 = i*z
        fadd    st4, st0
        fsubrp  st2, st0                        ; st0:st1 = 1-i*z | st2:st3 = 1+i*z
        mov     rdx, @@cdiv.enter
        call    rdx                             ; (1+i*z)/(1-i*z)
        mov     rdx, @@cln.enter
        call    rdx                             ; ln((1+i*z)/(1-i*z))
        fxch
        fchs                                    ; (-i)*ln((1+i*z)/(1-i*z))
        fldq    ??.0.5
        fmul    st2, st0
        fmulp                                   ; (1/2)*(-i)*ln((1+i*z)/(1-i*z))
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@catan.sse.fpu
$else
  @@catan.fpu
$end

; ------------------------------------------------------------------------------

; inverse complex cotangent
; z = a + ib
; cacotan(z)
; (1/2)*i*ln((i*z+1)/(i*z-1))
;
; rsp + 0 = a
; rsp + 8 = b
;
; return_1 = st1 = a'
; return_2 = st0 = b'

macro @@cacotan.sse.fpu
{
$usprocf @@cacotan

        movsd   xmm0, qword [rsp]               ; xmm0 = a
        movsd   xmm1, qword [rsp + 8]           ; xmm1 = b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:
        @@.fld.xmm0
        @@.fld.xmm1                             ; st0 = b | st1 = a
        ;
        fchs
        fxch                                    ; i*z = (ia-b)
        fld     st1
        fld     st1                             ; st0:st1 = i*z | st2:st3 = i*z
        fld1                                    ; st0 = 1 | st1:st2 = i*z | st3:st4 = i*z
        fadd    st4, st0
        fsubp   st2, st0                        ; st0:st1 = i*z-1 | st2:st3 = i*z+1
        mov     rdx, @@cdiv.enter.fpu
        call    rdx                             ; (i*z+1)/(i*z-1)
        mov     rdx, @@cln.enter.fpu
        call    rdx                             ; ln((i*z+1)/(i*z-1))
        fchs
        fxch                                    ; i*ln((i*z+1)/(i*z-1))
        fldq    ??.0.5
        fmul    st2, st0
        fmulp                                   ; (1/2)*i*ln((i*z+1)/(i*z-1))
   .return:
        @@.fstp.xmm1
        @@.fstp.xmm0
        retn
$endp
}

macro @@cacotan.fpu
{
$usprocf @@cacotan

        fld     qword [rsp]                     ; load a
        fld     qword [rsp + 8]                 ; load b
        @@discard                               ; optimizer directive - initial [xsp] values can be discarded
   .enter:                                      ; st0 = b | st1 = a
        fchs
        fxch                                    ; i*z = (ia-b)
        fld     st1
        fld     st1                             ; st0:st1 = i*z | st2:st3 = i*z
        fld1                                    ; st0 = 1 | st1:st2 = i*z | st3:st4 = i*z
        fadd    st4, st0
        fsubp   st2, st0                        ; st0:st1 = i*z-1 | st2:st3 = i*z+1
        mov     rdx, @@cdiv.enter
        call    rdx                             ; (i*z+1)/(i*z-1)
        mov     rdx, @@cln.enter
        call    rdx                             ; ln((i*z+1)/(i*z-1))
        fchs
        fxch                                    ; i*ln((i*z+1)/(i*z-1))
        fldq    ??.0.5
        fmul    st2, st0
        fmulp                                   ; (1/2)*i*ln((i*z+1)/(i*z-1))
   .return:
        retn
$endp
}

$ifdef _SSEX
  @@cacotan.sse.fpu
$else
  @@cacotan.fpu
$end

; ------------------------------------------------------------------------------
